import paho.mqtt.client as mqtt
import logging
import time
import socket

from holour import json_encode
from holour.msg import Status
from holour.mqtt.config import MQTTConfig


class MQTTClient:

    def __init__(self, name: str, mqtt_config: MQTTConfig, log_level=logging.INFO):
        self.log = logging.getLogger(name)
        self.log.setLevel(log_level)

        self.log.info(f"Setting up MQTT client...")
        self.ip = self._get_own_ip()
        self.connected = False
        self.mqtt_config = mqtt_config
        self.mqtt_client = mqtt.Client(client_id=mqtt_config.client_id)
        self.mqtt_client.on_connect = self._on_connect
        self.mqtt_client.on_disconnect = self._on_disconnect
        self.mqtt_client.on_message = self._on_message
        self.mqtt_client.will_set(self._status_topic(), self._status_payload(online=False), qos=2, retain=True)
        self.log.info("MQTT client setup is completed")

    def _get_own_ip(self) -> str:
        self.log.info("Retrieving own ip...")
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('8.8.8.8', 80))
        ip, port = s.getsockname()
        s.close()
        self.log.info(f"Running on: {ip}")
        return ip

    def _status_topic(self) -> str:
        return f'{self.mqtt_config.prefix}/status/{self.mqtt_config.client_id}'

    def _status_payload(self, online: bool) -> str:
        status = 'online' if online else 'offline'
        return json_encode(Status(self.mqtt_config.client_id, self.ip, status))

    def _connect(self):
        self.log.info(f"Connecting to MQTT broker on: {self.mqtt_config.address}:{self.mqtt_config.port}...")
        self.mqtt_client.connect(host=self.mqtt_config.address,
                                 port=self.mqtt_config.port,
                                 keepalive=self.mqtt_config.keep_alive)
        self.mqtt_client.loop_start()

    def _on_connect(self, client: mqtt.Client, userdata, flags, rc: int):
        self.log.info(f"Connected with result code: {rc}")
        self.connected = True
        client.publish(self._status_topic(), self._status_payload(online=True), qos=2, retain=True)

    def _on_disconnect(self, client: mqtt.Client, userdata, rc: int):
        self.log.info(f"Disconnected from MQTT-server")
        self.connected = False
        client.publish(self._status_topic(), self._status_payload(online=False), qos=2, retain=True)

    def _on_message(self, client: mqtt.Client, userdata, message: mqtt.MQTTMessage):
        raise NotImplementedError()

    def _wait_for_keyboard_interrupt(self):
        try:
            while True:
                time.sleep(10)
        except KeyboardInterrupt:
            self.mqtt_client.loop_stop()
            self.log.info("Bye bye")



import os
import yaml
import logging


def load_configuration(config_file: str, log: logging.Logger) -> dict:
    config = {}
    if not os.path.exists(config_file):
        log.error(f"The configuration file: {config_file} doesn't exist!")
        return config
    with open(config_file, 'r') as stream:
        try:
            config = yaml.safe_load(stream)
            log.info(f"Configuration: {config}")
        except yaml.YAMLError as e:
            log.error(e)
    return config





class Script:

    def __init__(self, filename: str, path: str, _type: str = ''):
        self._type = 'script'
        self.filename = filename
        self.path = path

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Script):
            return other.filename == self.filename and other.path == self.path
        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self):
        return f"Script<filename={self.filename}, path={self.path}>"


class WaypointConfig:

    def __init__(self, show_waypoints: bool, show_current_waypoint: bool, _type: str = ''):
        self._type = 'waypoint_config'
        self.show_waypoints = show_waypoints
        self.show_current_waypoint = show_current_waypoint

    def __eq__(self, other: object) -> bool:
        if isinstance(other, WaypointConfig):
            return other.show_waypoints == self.show_waypoints \
                   and other.show_current_waypoint == self.show_current_waypoint
        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self):
        return f"WaypointConfig<" \
               f"show_waypoints={self.show_waypoints}, show_current_waypoint={self.show_current_waypoint}>"



from unittest import TestCase
import logging
import os
from holour.mqtt import MQTTConfig
from holour import load_configuration


class Test(TestCase):

    def setUp(self) -> None:
        pass

    @staticmethod
    def path(filename: str) -> str:
        dir_path = os.path.dirname(os.path.realpath(__file__))
        dir_path = os.path.join(dir_path, 'config')
        return os.path.join(dir_path, filename)

    def test_mqtt_config(self):
        raw_config = {'client_id': 'test'}
        config = MQTTConfig(**raw_config)
        assert config.client_id == 'test'

    def test_load_configuration_without_file(self):
        with self.assertLogs(level=logging.ERROR) as log:
            output = load_configuration(self.path('doesnt.exist'), logging.getLogger('test'))

            assert output == {}
            self.assertEqual(len(log.output), 1)

    def test_load_configuration_invalid_file(self):
        with self.assertLogs(level=logging.ERROR) as log:
            output = load_configuration(self.path('load_invalid_config.yml'), logging.getLogger('test'))
            expected = {}

            assert output == {}, f"Expected: {expected}, got: {output}"
            self.assertEqual(len(log.output), 1)

    def test_load_configuration(self):
        with self.assertLogs(level=logging.INFO) as log:
            output = load_configuration(self.path('load_config_test.yml'), logging.getLogger('test'))
            expected = {'mqtt': {'client_id': 'test'}}

            assert output == expected, f"Expected: {expected}, got: {output}"
            self.assertEqual(len(log.output), 1)

    def test_load_configuration_and_init_config(self):
        output = load_configuration(self.path('load_full_config_test.yml'), logging.getLogger('test'))
        expected = {'mqtt': {
            'client_id': 'test_client',
            'address': '192.168.0.50',
            'port': 18831,
            'keep_alive': 30,
            'prefix': 'roberta/'}
        }
        assert output == expected, f"Expected: {expected}, got: {output}"

        config = MQTTConfig(**output['mqtt'])
        assert config.client_id == 'test_client'
        assert config.address == '192.168.0.50'



from unittest import TestCase

from holour import json_encode, json_decode
from holour.msg import Vector3, Pose, Poses


class Test(TestCase):

    def setUp(self) -> None:
        self.vector3 = Vector3(1, 2, 3)

        self.name = "waypoint_1"
        self.position = Vector3(1, 2, 3)
        self.orientation = Vector3(-1.1, -1.2, 1.3)
        self.pose = Pose(self.position, self.orientation, self.name)

        self.poses = Poses([])

    def test_vector3(self):
        vector3_string = json_encode(self.vector3)

        assert type(vector3_string) == str
        assert vector3_string == '{"_type": "vector3", "x": 1, "y": 2, "z": 3}', f"Got: {vector3_string}"

        vector3_decoded = json_decode(vector3_string)
        assert type(vector3_decoded) == Vector3
        assert vector3_decoded.x == 1
        assert vector3_decoded == self.vector3

    def test_vector3_equals(self):
        v2 = Vector3(1, 2, 3)
        v3 = Vector3(3000, 2, 3)

        assert self.vector3 == v2
        assert self.vector3 != v3
        assert self.vector3 != "hej"

    def test_vector3_repr(self):
        expected, got = 'x=1', f'{self.vector3}'
        assert expected in got, f"Expected {expected} in got: {got}"

    def test_pose(self):
        pose_string = json_encode(self.pose)

        assert type(pose_string) == str
        assert pose_string == '{"_type": "pose", "name": "waypoint_1", ' \
                              '"position": {"_type": "vector3", "x": 1, "y": 2, "z": 3}, ' \
                              '"orientation": {"_type": "vector3", "x": -1.1, "y": -1.2, "z": 1.3}}'

        pose_decoded: Pose = json_decode(pose_string)
        assert type(pose_decoded) == Pose, f"Expected object of type {Pose}"
        assert pose_decoded.name == self.name
        assert pose_decoded.position == self.position
        assert pose_decoded.orientation == self.orientation
        assert pose_decoded == self.pose

    def test_pose_equals(self):
        pose2 = Pose(Vector3(1, 2, 3), Vector3(-1.1, -1.2, 1.3), "waypoint_1")
        pose3 = Pose(Vector3(1, 2, 3), Vector3(-1.1, -1.2, 1.3), "some name")

        assert self.pose == pose2
        assert self.pose != pose3
        assert self.pose != "hej"

    def test_pose_repr(self):
        expected, got = f'name={self.name}', f'{self.pose}'
        assert expected in got, f"Expected {expected} in got: {got}"

    def test_poses(self):
        list_of_poses = [
            Pose(Vector3(1, 1, 1), Vector3(-1.1, -1.2, 1.3), "waypoint_1"),
            Pose(Vector3(2, 2, 2), Vector3(-1.2, -1.1, 1.3), "waypoint_2"),
            Pose(Vector3(3, 3, 3), Vector3(-1.3, -1.2, 1.1), "waypoint_3"),
            Pose(Vector3(3, 3, 3), Vector3(-1.3, -1.2, 1.1), "waypoint_4"),
        ]
        poses = Poses(list_of_poses)

        poses_string = json_encode(poses)
        assert type(poses_string) == str
        assert "connected" in poses_string
        assert "poses" in poses_string

        poses_decoded: Poses = json_decode(poses_string)
        assert type(poses_decoded) == Poses, f"Got: {type(poses_decoded)}. Expected {Poses}"
        assert len(poses_decoded.poses) == len(list_of_poses)
        assert poses_decoded.poses == list_of_poses
        assert poses_decoded == poses, "The decoded object must be equal to the encoded"

    def test_poses_equals(self):
        list_of_poses = [
            Pose(Vector3(1, 1, 1), Vector3(-1.1, -1.2, 1.3), "waypoint_1"),
            Pose(Vector3(2, 2, 2), Vector3(-1.2, -1.1, 1.3), "waypoint_2"),
        ]
        poses1 = Poses(list_of_poses)
        poses2 = Poses(list_of_poses)
        poses3 = Poses(list_of_poses, connected=True)

        assert poses1 == poses2
        assert poses1 != poses3
        assert poses1 != list_of_poses

    def test_poses_add(self):
        assert len(self.poses.poses) == 0

        self.poses.add(self.pose)
        assert len(self.poses.poses) == 1
        assert self.pose in self.poses.poses

    def test_poses_repr(self):
        expected, got = f'connected=False', f'{self.poses}'
        assert expected in got, f"Expected {expected} in got: {got}"



from unittest import TestCase

from holour import json_encode, json_decode
from holour.msg import Status


class Test(TestCase):

    def test_status(self):
        status = Status('data_objects', '127.0.0.1', Status.OFFLINE)
        status_string = json_encode(status)
        expected_string = '{"_type": "status", "id": "data_objects", "ip": "127.0.0.1", "status": "offline"}'

        assert type(status_string) == str
        assert status_string == expected_string, f"Expected {expected_string}, got: {status_string}"

        status_decoded = json_decode(status_string)
        assert type(status_decoded) == Status, f"Got: {type(status_decoded)}. Expected {Status}"
        assert status_decoded == status, "The decoded object must be equal to the encoded"

    def test_status_equals(self):
        s1 = Status('data_objects', '127.0.0.1', Status.OFFLINE)
        s2 = Status('data_objects', '127.0.0.1', Status.OFFLINE)
        s3 = Status('data_objects', '127.0.0.1', Status.ONLINE)

        assert s1 == s2
        assert s1 != s3
        assert s1 != "not status"

    def test_status_repr(self):
        status = Status('data_objects', '127.0.0.1', Status.OFFLINE)
        expected, got = '127.0.0.1', f'{status}'

        assert expected in got, f"Expected {expected} in got: {got}"



import logging
import sys


def setup_log() -> None:
    log_format = logging.Formatter('[%(asctime)s] [%(levelname)s] [%(name)s:%(funcName)s] - %(message)s')
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(log_format)
    logging.basicConfig(level=logging.DEBUG, handlers=[handler])



from typing import List


class Vector3:

    def __init__(self, x: float, y: float, z: float, _type: str = ''):
        self._type = 'vector3'
        self.x = x
        self.y = y
        self.z = z

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Vector3):
            return other.x == self.x and other.y == self.y and other.z == self.z
        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self):
        return f"Vector3<x={self.x}, y={self.y}, z={self.z}>"


class Pose:

    def __init__(self, position: Vector3, orientation: Vector3, name: str = '', _type: str = ''):
        self._type = 'pose'
        self.name = name
        self.position = position
        self.orientation = orientation

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Pose):
            return other.name == self.name and other.position == self.position and other.orientation == self.orientation
        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self):
        return f"Pose<name={self.name}, position={self.position}, orientation={self.orientation}>"


class Poses:

    def __init__(self, poses: List[Pose], connected: bool = False, _type: str = ''):
        self._type = 'poses'
        self.connected = connected
        self.poses = poses

    def add(self, pose: Pose):
        self.poses.append(pose)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Poses):
            return other.connected == self.connected and other.poses == self.poses
        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self):
        return f"Poses<connected={self.connected}, poses={self.poses}>"
